# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import re


def get_property(prop):
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open("swane/__init__.py").read()
    )
    return result.group(1)


setup(
    name="swane",
    version=get_property("__version__"),
    description="Standardized Workflow for Advanced Neuroimaging in Epilepsy",
    author="LICE - Commissione Neuroimmagini",
    author_email="dev@lice.it",
    packages=find_packages(exclude=["swane.tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
    ],
    license="MIT",
    install_requires=[
        "networkx==3.4.2",
        "nipype==1.10.0",
        "PySide6",
        "pydicom==3.0.1",
        "configparser<=7.1.0",
        "psutil==7.0.0",
        "swane_supplement>=0.1.2",
        "matplotlib==3.10.1",
        "nibabel>=5.3.0,<6",
        "packaging",
        "PySide6_VerticalQTabWidget==0.0.3",
        "numpy==2.2.4",
        "cryptography",
        "dicom-sequence-classifier==1.0.4",
    ],
    python_requires=">=3.10",
    entry_points={"gui_scripts": ["swane = swane.__main__:main"]},
)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

import shutil
from nipype.interfaces.fsl import SwapDimensions
from os.path import abspath
import os
from nipype.interfaces.base import (
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
    isdefined,
)
from swane.nipype_pipeline.nodes.Orient import Orient


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class ForceOrientInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="the input image")
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class ForceOrientOutputSpec(TraitedSpec):
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class ForceOrient(BaseInterface):
    """
    Converts an image in radiological convention and in RL PA IS orientation.

    """

    input_spec = ForceOrientInputSpec
    output_spec = ForceOrientOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file = self._gen_outfilename()
        shutil.copy(self.inputs.in_file, self.inputs.out_file)
        get_orient = Orient(in_file=self.inputs.out_file)
        get_orient.inputs.get_orient = True
        res = get_orient.run()
        if res.outputs.orient == "NEUROLOGICAL":
            swap_nr = SwapDimensions()
            swap_nr.inputs.in_file = self.inputs.out_file
            swap_nr.inputs.out_file = self.inputs.out_file
            swap_nr.inputs.new_dims = ("-x", "y", "z")
            swap_nr.run()
            swap_orient = Orient(in_file=self.inputs.out_file)
            swap_orient.inputs.swap_orient = True
            swap_orient.run()
        swap_dim = SwapDimensions()
        swap_dim.inputs.in_file = self.inputs.out_file
        swap_dim.inputs.out_file = self.inputs.out_file
        swap_dim.inputs.new_dims = ("RL", "PA", "IS")
        swap_dim.run()

        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = os.path.basename(self.inputs.in_file)
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        return outputs



from datetime import datetime
import os
from multiprocessing import cpu_count
from os.path import abspath

import swane_supplement
from swane.config.ConfigManager import ConfigManager
from swane.utils.SubjectInputStateList import SubjectInputStateList
from swane.utils.DataInputList import DataInputList as DIL, FMRI_NUM
from swane.config.config_enums import (
    PLANES,
    CORE_LIMIT,
    BLOCK_DESIGN,
    GlobalPrefCategoryList,
)
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.workflows.linear_reg_workflow import linear_reg_workflow
from swane.nipype_pipeline.workflows.task_fMRI_workflow import task_fMRI_workflow
from swane.nipype_pipeline.workflows.nonlinear_reg_workflow import (
    nonlinear_reg_workflow,
)
from swane.nipype_pipeline.workflows.ref_workflow import ref_workflow
from swane.nipype_pipeline.workflows.freesurfer_workflow import freesurfer_workflow
from swane.nipype_pipeline.workflows.flat1_workflow import flat1_workflow
from swane.nipype_pipeline.workflows.func_map_workflow import func_map_workflow
from swane.nipype_pipeline.workflows.venous_workflow import venous_workflow
from swane.nipype_pipeline.workflows.dti_preproc_workflow import dti_preproc_workflow
from swane.nipype_pipeline.workflows.tractography_workflow import (
    tractography_workflow,
    SIDES,
)
from swane.config.preference_list import TRACTS
from swane.utils.DependencyManager import DependencyManager
from swane.nipype_pipeline.engine.MonitoredMultiProcPlugin import (
    MonitoredMultiProcPlugin,
)


DEBUG = False


# TODO implementazione error manager
class MainWorkflow(CustomWorkflow):
    Result_DIR: str = "results"

    is_resource_monitor: bool = False
    max_cpu: int = -1
    max_gpu: int = -1
    multicore_node_limit: CORE_LIMIT = CORE_LIMIT.SOFT_CAP

    name: str
    base_dir: str
    global_config: ConfigManager
    subject_config: ConfigManager
    dependency_manager: DependencyManager
    subject_input_state_list: SubjectInputStateList

    def __init__(
        self,
        name: str,
        base_dir: str,
        global_config: ConfigManager,
        subject_config: ConfigManager,
        dependency_manager: DependencyManager,
        subject_input_state_list: SubjectInputStateList,
    ):
        """
        Create the Workflows and their sub-workflows based on the list of available data inputs

        Parameters
        ----------
        name : str
        base_dir : str
        global_config : ConfigManager
            The app global configurations.
        subject_config : ConfigManager
            The subject specific configurations.
        dependency_manager: DependencyManager
            The state of application dependency
        subject_input_state_list : SubjectInputStateList
            The list of all available input data from the DICOM directory.

        """

        super().__init__(name, base_dir)

        self.global_config = global_config
        self.subject_config = subject_config
        self.dependency_manager = dependency_manager
        self.subject_input_state_list = subject_input_state_list

        if not subject_input_state_list.is_ref_loaded:
            return

        self.set_resources_configuration()
        self.set_analyses_request()

        self.launch_3dt1_analysis()
        self.launch_ai_analysis()
        self.launch_freesurfer_analysis()
        self.launch_3dflair_analysis()
        self.launch_flat1_analysis()
        self.launch_2dflair_analysis()
        self.launch_t2cor_analysis()
        self.launch_mdc_analysis()
        self.launch_asl_analysis()
        self.launch_pet_analysis()
        self.launch_venous_analysis()
        self.launch_dti_analysis()
        self.launch_fMRI_analysis()

        # Remove reference to original variables to prevent crash during subprocess spawn in MacOS
        # Maybe this can be solved setting fork subprocess method too
        self.global_config = None
        self.subject_config = None
        self.dependency_manager = None
        self.subject_input_state_list = None

    def set_resources_configuration(self):
        # CPU cores and memory management
        self.is_resource_monitor = self.global_config.getboolean_safe(
            GlobalPrefCategoryList.PERFORMANCE, "resource_monitor"
        )
        self.max_cpu = self.global_config.getint_safe(
            GlobalPrefCategoryList.PERFORMANCE, "max_subj_cu"
        )
        if self.max_cpu < 1:
            self.max_cpu = cpu_count()
        self.multicore_node_limit = self.global_config.getenum_safe(
            GlobalPrefCategoryList.PERFORMANCE, "multicore_node_limit"
        )
        # GPU management
        self.max_gpu = self.global_config.getint_safe(
            GlobalPrefCategoryList.PERFORMANCE, "max_subj_gpu"
        )
        if self.max_gpu < 0:
            self.max_gpu = MonitoredMultiProcPlugin.gpu_count()

        try:
            if not self.dependency_manager.is_cuda():
                self.subject_config[DIL.DTI]["cuda"] = "false"
            else:
                self.subject_config[DIL.DTI]["cuda"] = self.global_config[
                    GlobalPrefCategoryList.PERFORMANCE
                ]["cuda"]
        except:
            self.subject_config[DIL.DTI]["cuda"] = "false"

        self.subject_config.sections()

        # TODO - NOT USED, WHY?
        max_node_cpu = max(int(self.max_cpu / 2), 1)

    def set_analyses_request(self):
        # Check for FreeSurfer requirement and request
        self.is_freesurfer = (
            self.dependency_manager.is_freesurfer()
            and self.subject_config.get_workflow_freesurfer_pref()
        )
        self.is_hippo_amyg_labels = (
            self.dependency_manager.is_freesurfer_matlab()
            and self.subject_config.get_workflow_hippo_pref()
        )

        # Check for FLAT1 requirement and request
        self.is_flat1 = (
            self.subject_config.getboolean_safe(DIL.T13D, "flat1")
            and self.subject_input_state_list[DIL.FLAIR3D].loaded
        )
        # Check for Asymmetry Index request
        self.is_ai = (
            self.subject_config.getboolean_safe(DIL.PET, "ai")
            and self.subject_input_state_list[DIL.PET].loaded
        ) or (
            self.subject_config.getboolean_safe(DIL.ASL, "ai")
            and self.subject_input_state_list[DIL.ASL].loaded
        )
        # Check for Tractography request
        self.is_tractography = self.subject_config.getboolean_safe(
            DIL.DTI, "tractography"
        )

    def launch_3dt1_analysis(self):
        ref_dir = self.subject_input_state_list.get_dicom_dir(DIL.T13D)
        self.t1 = ref_workflow(
            name=DIL.T13D.value.workflow_name,
            dicom_dir=ref_dir,
            config=self.subject_config[DIL.T13D],
        )
        self.t1.long_name = "3D T1w analysis"
        self.add_nodes([self.t1])

        self.t1.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="ref",
            sub_folder=self.Result_DIR,
        )
        self.t1.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="ref_brain",
            sub_folder=self.Result_DIR,
        )

    def launch_ai_analysis(self):
        if not self.is_ai:
            return

        # Non linear registration for Asymmetry Index
        self.sym = nonlinear_reg_workflow(name="sym")
        self.sym.long_name = "Symmetric atlas registration"

        sym_inputnode = self.sym.get_node("inputnode")
        sym_template = swane_supplement.sym_template
        sym_inputnode.inputs.atlas = sym_template
        self.connect(self.t1, "outputnode.ref_brain", self.sym, "inputnode.in_file")

    def launch_freesurfer_analysis(self):
        if not self.is_freesurfer:
            return

        # FreeSurfer analysis
        self.freesurfer = freesurfer_workflow(
            name="freesurfer",
            is_hippo_amyg_labels=self.is_hippo_amyg_labels,
            max_cpu=self.max_cpu,
            multicore_node_limit=self.multicore_node_limit,
        )
        self.freesurfer.long_name = "Freesurfer analysis"

        freesurfer_inputnode = self.freesurfer.get_node("inputnode")
        freesurfer_inputnode.inputs.subjects_dir = self.base_dir
        self.connect(self.t1, "outputnode.ref", self.freesurfer, "inputnode.ref")

        self.freesurfer.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="pial",
            sub_folder=self.Result_DIR,
        )
        self.freesurfer.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="white",
            sub_folder=self.Result_DIR,
        )
        self.freesurfer.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="vol_label_file",
            sub_folder=self.Result_DIR,
        )
        if self.is_hippo_amyg_labels:
            regex_subs = [("-T1.*.mgz", ".mgz")]
            self.freesurfer.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="lh_hippoAmygLabels",
                sub_folder="scene.segmentHA",
                regexp_substitutions=regex_subs,
            )
            self.freesurfer.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="rh_hippoAmygLabels",
                sub_folder="scene.segmentHA",
                regexp_substitutions=regex_subs,
            )

    def launch_3dflair_analysis(self):
        if not self.subject_input_state_list[DIL.FLAIR3D].loaded:
            return

        # 3D Flair analysis
        flair_dir = self.subject_input_state_list.get_dicom_dir(DIL.FLAIR3D)
        self.flair = linear_reg_workflow(
            name=DIL.FLAIR3D.value.workflow_name,
            dicom_dir=flair_dir,
            config=self.subject_config[DIL.FLAIR3D],
        )
        self.flair.long_name = "3D Flair analysis"
        self.add_nodes([self.flair])

        flair_inputnode = self.flair.get_node("inputnode")
        flair_inputnode.inputs.output_name = "flair"
        self.connect(self.t1, "outputnode.ref_brain", self.flair, "inputnode.reference")

        self.flair.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="registered_file",
            sub_folder=self.Result_DIR,
        )

        self.flair.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="betted_registered_file",
            sub_folder=self.Result_DIR,
        )

        # TODO: explore possibility of freesurfer based asymmetry index
        # if is_freesurfer:
        #     from swane.nipype_pipeline.workflows.freesurfer_asymmetry_index_workflow import freesurfer_asymmetry_index_workflow
        #     flair_ai = freesurfer_asymmetry_index_workflow(name="flair_ai")
        #     self.connect(flair, "outputnode.registered_file", flair_ai, "inputnode.in_file")
        #     self.connect(freesurfer, "outputnode.vol_label_file_nii", flair_ai, "inputnode.seg_file")

    def launch_flat1_analysis(self):
        if not self.is_flat1:
            return

        # Non linear registration to MNI1mm Atlas for FLAT1
        self.mni1 = nonlinear_reg_workflow(name="mni1")
        self.mni1.long_name = "MNI atlas registration"

        mni1_inputnode = self.mni1.get_node("inputnode")
        mni1_path = abspath(
            os.path.join(
                os.environ["FSLDIR"], "data/standard/MNI152_T1_1mm_brain.nii.gz"
            )
        )
        mni1_inputnode.inputs.atlas = mni1_path
        self.connect(self.t1, "outputnode.ref_brain", self.mni1, "inputnode.in_file")

        # FLAT1 analysis
        self.flat1 = flat1_workflow(name="FLAT1", mni1_dir=mni1_path)
        self.flat1.long_name = "FLAT1 analysis"

        self.connect(self.t1, "outputnode.ref_brain", self.flat1, "inputnode.ref_brain")
        self.connect(
            self.flair,
            "outputnode.registered_file",
            self.flat1,
            "inputnode.flair_brain",
        )
        self.connect(
            self.mni1,
            "outputnode.fieldcoeff_file",
            self.flat1,
            "inputnode.ref_2_mni1_warp",
        )
        self.connect(
            self.mni1,
            "outputnode.inverse_warp",
            self.flat1,
            "inputnode.ref_2_mni1_inverse_warp",
        )

        self.flat1.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="extension_z",
            sub_folder=self.Result_DIR,
        )
        self.flat1.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="junction_z",
            sub_folder=self.Result_DIR,
        )
        self.flat1.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="binary_flair",
            sub_folder=self.Result_DIR,
        )

    def launch_2dflair_analysis(self):
        for plane in PLANES:
            if (
                DIL["FLAIR2D_%s" % plane.name] in self.subject_input_state_list
                and self.subject_input_state_list[DIL["FLAIR2D_%s" % plane.name]].loaded
            ):
                flair_dir = self.subject_input_state_list.get_dicom_dir(
                    DIL["FLAIR2D_%s" % plane.name]
                )
                self.flair2d = linear_reg_workflow(
                    name=DIL["FLAIR2D_%s" % plane.name].value.workflow_name,
                    dicom_dir=flair_dir,
                    config=None,
                    is_volumetric=False,
                )
                self.flair2d.long_name = "2D %s FLAIR analysis" % plane.value
                self.add_nodes([self.flair2d])

                flair2d_inputnode = self.flair2d.get_node("inputnode")
                flair2d_inputnode.inputs.output_name = "flair2d_%s" % plane
                self.connect(
                    self.t1, "outputnode.ref_brain", self.flair2d, "inputnode.reference"
                )

                self.flair2d.sink_result(
                    save_path=self.base_dir,
                    result_node="outputnode",
                    result_name="betted_registered_file",
                    sub_folder=self.Result_DIR,
                )

    def launch_t2cor_analysis(self):
        if (
            not DIL.T2_COR in self.subject_input_state_list
            or not self.subject_input_state_list[DIL.T2_COR].loaded
        ):
            return

        t2_cor_dir = self.subject_input_state_list.get_dicom_dir(DIL.T2_COR)
        self.t2_cor = linear_reg_workflow(
            name=DIL.T2_COR.value.workflow_name,
            dicom_dir=t2_cor_dir,
            config=None,
            is_volumetric=False,
            is_partial_coverage=True,
        )
        self.t2_cor.long_name = "2D coronal T2 analysis"
        self.add_nodes([self.t2_cor])

        t2_cor_inputnode = self.t2_cor.get_node("inputnode")
        t2_cor_inputnode.inputs.output_name = "t2_cor"
        self.connect(self.t1, "outputnode.ref", self.t2_cor, "inputnode.reference")
        self.connect(
            self.t1, "outputnode.ref_mask", self.t2_cor, "inputnode.brain_mask"
        )

        self.t2_cor.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="registered_file",
            sub_folder=self.Result_DIR,
        )

        self.t2_cor.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="betted_registered_file",
            sub_folder=self.Result_DIR,
        )

    def launch_mdc_analysis(self):
        if not self.subject_input_state_list[DIL.MDC].loaded:
            return

        # MDC analysis
        mdc_dir = self.subject_input_state_list.get_dicom_dir(DIL.MDC)
        self.mdc = linear_reg_workflow(
            name=DIL.MDC.value.workflow_name,
            dicom_dir=mdc_dir,
            config=self.subject_config[DIL.MDC],
        )
        self.mdc.long_name = "Post-contrast 3D T1w analysis"
        self.add_nodes([self.mdc])

        mdc_inputnode = self.mdc.get_node("inputnode")
        mdc_inputnode.inputs.output_name = "mdc"
        self.connect(self.t1, "outputnode.ref_brain", self.mdc, "inputnode.reference")

        self.mdc.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="registered_file",
            sub_folder=self.Result_DIR,
        )

        self.mdc.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="betted_registered_file",
            sub_folder=self.Result_DIR,
        )

    def launch_asl_analysis(self):
        if not self.subject_input_state_list[DIL.ASL].loaded:
            return

        # ASL analysis
        asl_dir = self.subject_input_state_list.get_dicom_dir(DIL.ASL)
        self.asl = func_map_workflow(
            name=DIL.ASL.value.workflow_name,
            dicom_dir=asl_dir,
            is_freesurfer=self.is_freesurfer,
            config=self.subject_config[DIL.ASL],
        )
        self.asl.long_name = "Arterial Spin Labelling analysis"

        self.connect(self.t1, "outputnode.ref_brain", self.asl, "inputnode.reference")
        self.connect(self.t1, "outputnode.ref_mask", self.asl, "inputnode.brain_mask")

        self.asl.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="registered_file",
            sub_folder=self.Result_DIR,
        )

        if self.is_freesurfer:
            self.connect(
                self.freesurfer,
                "outputnode.subjects_dir",
                self.asl,
                "inputnode.freesurfer_subjects_dir",
            )
            self.connect(
                self.freesurfer,
                "outputnode.subject_id",
                self.asl,
                "inputnode.freesurfer_subject_id",
            )
            self.connect(
                self.freesurfer, "outputnode.bgROI", self.asl, "inputnode.bgROI"
            )

            self.asl.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="surf_lh",
                sub_folder=self.Result_DIR,
            )
            self.asl.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="surf_rh",
                sub_folder=self.Result_DIR,
            )
            self.asl.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="zscore",
                sub_folder=self.Result_DIR,
            )
            self.asl.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="zscore_surf_lh",
                sub_folder=self.Result_DIR,
            )
            self.asl.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="zscore_surf_rh",
                sub_folder=self.Result_DIR,
            )

        if self.subject_config.getboolean_safe(DIL.ASL, "ai"):
            self.connect(
                self.sym,
                "outputnode.fieldcoeff_file",
                self.asl,
                "inputnode.ref_2_sym_warp",
            )
            self.connect(
                self.sym,
                "outputnode.inverse_warp",
                self.asl,
                "inputnode.ref_2_sym_invwarp",
            )

            self.asl.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="ai",
                sub_folder=self.Result_DIR,
            )

            if self.is_freesurfer:
                self.asl.sink_result(
                    save_path=self.base_dir,
                    result_node="outputnode",
                    result_name="ai_surf_lh",
                    sub_folder=self.Result_DIR,
                )
                self.asl.sink_result(
                    save_path=self.base_dir,
                    result_node="outputnode",
                    result_name="ai_surf_rh",
                    sub_folder=self.Result_DIR,
                )

    def launch_pet_analysis(self):
        if not self.subject_input_state_list[
            DIL.PET
        ].loaded:  # and check_input['ct_brain']:
            return

        # PET analysis
        pet_dir = self.subject_input_state_list.get_dicom_dir(DIL.PET)
        self.pet = func_map_workflow(
            name=DIL.PET.value.workflow_name,
            dicom_dir=pet_dir,
            is_freesurfer=self.is_freesurfer,
            config=self.subject_config[DIL.PET],
        )
        self.pet.long_name = "Pet analysis"

        self.connect(self.t1, "outputnode.ref", self.pet, "inputnode.reference")
        self.connect(self.t1, "outputnode.ref_mask", self.pet, "inputnode.brain_mask")

        self.pet.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="registered_file",
            sub_folder=self.Result_DIR,
        )

        if self.is_freesurfer:
            self.connect(
                self.freesurfer,
                "outputnode.subjects_dir",
                self.pet,
                "inputnode.freesurfer_subjects_dir",
            )
            self.connect(
                self.freesurfer,
                "outputnode.subject_id",
                self.pet,
                "inputnode.freesurfer_subject_id",
            )
            self.connect(
                self.freesurfer, "outputnode.bgROI", self.pet, "inputnode.bgROI"
            )

            self.pet.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="surf_lh",
                sub_folder=self.Result_DIR,
            )
            self.pet.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="surf_rh",
                sub_folder=self.Result_DIR,
            )
            self.pet.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="zscore",
                sub_folder=self.Result_DIR,
            )
            self.pet.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="zscore_surf_lh",
                sub_folder=self.Result_DIR,
            )
            self.pet.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="zscore_surf_rh",
                sub_folder=self.Result_DIR,
            )

            # TODO work in progress for segmentation based asymmetry study
            # from swane.nipype_pipeline.workflows.freesurfer_asymmetry_index_workflow import freesurfer_asymmetry_index_workflow
            # pet_ai = freesurfer_asymmetry_index_workflow(name="pet_ai")
            # self.connect(pet, "outputnode.registered_file", pet_ai, "inputnode.in_file")
            # self.connect(freesurfer, "outputnode.vol_label_file_nii", pet_ai, "inputnode.seg_file")

        if self.subject_config.getboolean_safe(DIL.PET, "ai"):
            self.connect(
                self.sym,
                "outputnode.fieldcoeff_file",
                self.pet,
                "inputnode.ref_2_sym_warp",
            )
            self.connect(
                self.sym,
                "outputnode.inverse_warp",
                self.pet,
                "inputnode.ref_2_sym_invwarp",
            )

            self.pet.sink_result(
                save_path=self.base_dir,
                result_node="outputnode",
                result_name="ai",
                sub_folder=self.Result_DIR,
            )

            if self.is_freesurfer:
                self.pet.sink_result(
                    save_path=self.base_dir,
                    result_node="outputnode",
                    result_name="ai_surf_lh",
                    sub_folder=self.Result_DIR,
                )
                self.pet.sink_result(
                    save_path=self.base_dir,
                    result_node="outputnode",
                    result_name="ai_surf_rh",
                    sub_folder=self.Result_DIR,
                )

    def launch_venous_analysis(self):
        if (
            not self.subject_input_state_list[DIL.VENOUS].loaded
            or not self.subject_input_state_list[DIL.VENOUS].volumes
            + self.subject_input_state_list[DIL.VENOUS2].volumes
            == 2
        ):
            return

        # Venous analysis
        venous_dir = self.subject_input_state_list.get_dicom_dir(DIL.VENOUS)
        venous2_dir = None
        if self.subject_input_state_list[DIL.VENOUS2].loaded:
            venous2_dir = self.subject_input_state_list.get_dicom_dir(DIL.VENOUS2)
        self.venous = venous_workflow(
            DIL.VENOUS.value.workflow_name,
            venous_dir,
            self.subject_config[DIL.VENOUS],
            venous2_dir,
        )
        self.venous.long_name = "Venous MRA analysis"

        self.connect(
            self.t1, "outputnode.ref_brain", self.venous, "inputnode.ref_brain"
        )

        self.venous.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="veins",
            sub_folder=self.Result_DIR,
        )

    def launch_dti_analysis(self):
        if not self.subject_input_state_list[DIL.DTI].loaded:
            return

        # DTI analysis
        dti_dir = self.subject_input_state_list.get_dicom_dir(DIL.DTI)
        mni_dir = abspath(
            os.path.join(
                os.environ["FSLDIR"], "data/standard/MNI152_T1_2mm_brain.nii.gz"
            )
        )

        self.dti_preproc = dti_preproc_workflow(
            name=DIL.DTI.value.workflow_name,
            dti_dir=dti_dir,
            config=self.subject_config[DIL.DTI],
            mni_dir=mni_dir,
            max_cpu=self.max_cpu,
            multicore_node_limit=self.multicore_node_limit,
        )
        self.dti_preproc.long_name = "Diffusion Tensor Imaging preprocessing"
        self.connect(
            self.t1, "outputnode.ref_brain", self.dti_preproc, "inputnode.ref_brain"
        )

        self.dti_preproc.sink_result(
            save_path=self.base_dir,
            result_node="outputnode",
            result_name="FA",
            sub_folder=self.Result_DIR,
        )

        if self.is_tractography:
            for tract in TRACTS.keys():
                try:
                    if not self.subject_config.getboolean_safe(DIL.DTI, tract):
                        continue
                except:
                    continue

                tract_workflow = tractography_workflow(
                    tract, self.subject_config[DIL.DTI]
                )
                if tract_workflow is not None:
                    tract_workflow.long_name = TRACTS[tract][0] + " tractography"
                    self.connect(
                        self.dti_preproc,
                        "outputnode.fsamples",
                        tract_workflow,
                        "inputnode.fsamples",
                    )
                    self.connect(
                        self.dti_preproc,
                        "outputnode.nodiff_mask_file",
                        tract_workflow,
                        "inputnode.mask",
                    )
                    self.connect(
                        self.dti_preproc,
                        "outputnode.phsamples",
                        tract_workflow,
                        "inputnode.phsamples",
                    )
                    self.connect(
                        self.dti_preproc,
                        "outputnode.thsamples",
                        tract_workflow,
                        "inputnode.thsamples",
                    )
                    self.connect(
                        self.t1,
                        "outputnode.ref_brain",
                        tract_workflow,
                        "inputnode.ref_brain",
                    )
                    self.connect(
                        self.dti_preproc,
                        "outputnode.diff2ref_mat",
                        tract_workflow,
                        "inputnode.diff2ref_mat",
                    )
                    self.connect(
                        self.dti_preproc,
                        "outputnode.ref2diff_mat",
                        tract_workflow,
                        "inputnode.ref2diff_mat",
                    )
                    self.connect(
                        self.dti_preproc,
                        "outputnode.mni2ref_warp",
                        tract_workflow,
                        "inputnode.mni2ref_warp",
                    )

                    for side in SIDES:
                        tract_workflow.sink_result(
                            save_path=self.base_dir,
                            result_node="outputnode",
                            result_name="waytotal_%s" % side,
                            sub_folder=self.Result_DIR + ".dti",
                        )
                        tract_workflow.sink_result(
                            save_path=self.base_dir,
                            result_node="outputnode",
                            result_name="fdt_paths_%s" % side,
                            sub_folder=self.Result_DIR + ".dti",
                        )

    def launch_fMRI_analysis(self):
        # Check for Task FMRI sequences
        for y in range(FMRI_NUM):

            if not self.subject_input_state_list[DIL["FMRI_%d" % y]].loaded:
                continue

            dicom_dir = self.subject_input_state_list.get_dicom_dir(DIL["FMRI_%d" % y])
            self.fMRI = task_fMRI_workflow(
                name=DIL["FMRI_%d" % y].value.workflow_name,
                dicom_dir=dicom_dir,
                config=self.subject_config[DIL["FMRI_%d" % y]],
                base_dir=self.base_dir,
            )
            self.fMRI.long_name = "Task fMRI analysis - %d" % y
            self.connect(
                self.t1, "outputnode.ref_brain", self.fMRI, "inputnode.ref_BET"
            )
            for thresh_i in range(1, 4):
                self.fMRI.sink_result(
                    save_path=self.base_dir,
                    result_node="outputnode",
                    result_name="threshold_file_cont1_thresh%d" % thresh_i,
                    sub_folder=self.Result_DIR + ".fMRI",
                )
                if (
                    self.subject_config.getenum_safe(DIL["FMRI_%d" % y], "block_design")
                    == BLOCK_DESIGN.RARB
                ):
                    self.fMRI.sink_result(
                        save_path=self.base_dir,
                        result_node="outputnode",
                        result_name="threshold_file_cont2_thresh%d" % thresh_i,
                        sub_folder=self.Result_DIR + ".fMRI",
                    )



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

import shutil
from nipype.interfaces.base import InputMultiObject
from nipype.interfaces.fsl import ImageStats
from os.path import abspath
from swane.config.config_enums import VEIN_DETECTION_MODE
from nipype.interfaces.base import traits
from nipype.interfaces.base import (
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class VenousCheckInputSpec(BaseInterfaceInputSpec):
    in_files = InputMultiObject(File(exists=True), desc="List of splitted file")
    detection_mode = traits.Enum(VEIN_DETECTION_MODE, argstr="-m %d", usedefault=True)
    out_file_veins = File(desc="the output venous image")
    out_file_anat = File(desc="the output anatomic image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class VenousCheckOutputSpec(TraitedSpec):
    out_file_veins = File(desc="the output venous image")
    out_file_anat = File(desc="the output anatomic image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class VenousCheck(BaseInterface):
    """
    Recognises the venous phase from the anatomic image of a phase contrast sequence based criteria specified by user.

    """

    input_spec = VenousCheckInputSpec
    output_spec = VenousCheckOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file_veins = abspath("veins.nii.gz")
        self.inputs.out_file_anat = abspath("veins_anat.nii.gz")

        if self.inputs.detection_mode == VEIN_DETECTION_MODE.FIRST:
            # Always first
            shutil.copy(self.inputs.in_files[0], self.inputs.out_file_veins)
            shutil.copy(self.inputs.in_files[1], self.inputs.out_file_anat)
            return runtime
        elif self.inputs.detection_mode == VEIN_DETECTION_MODE.SECOND:
            # Always second
            shutil.copy(self.inputs.in_files[1], self.inputs.out_file_veins)
            shutil.copy(self.inputs.in_files[0], self.inputs.out_file_anat)
            return runtime
        elif self.inputs.detection_mode == VEIN_DETECTION_MODE.MEAN:
            # Mean value mode
            op_string = "-M"
        else:
            # Standard deviation mode (default)
            op_string = "-s"

        s0 = ImageStats()
        s0.inputs.in_file = self.inputs.in_files[0]
        s0.inputs.op_string = op_string
        res0 = s0.run()
        s1 = ImageStats()
        s1.inputs.in_file = self.inputs.in_files[1]
        s1.inputs.op_string = op_string
        res1 = s1.run()
        if res0.outputs.out_stat < res1.outputs.out_stat:
            shutil.copy(self.inputs.in_files[0], self.inputs.out_file_veins)
            shutil.copy(self.inputs.in_files[1], self.inputs.out_file_anat)
        else:
            shutil.copy(self.inputs.in_files[1], self.inputs.out_file_veins)
            shutil.copy(self.inputs.in_files[0], self.inputs.out_file_anat)

        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file_veins"] = abspath("veins.nii.gz")
        outputs["out_file_anat"] = abspath("veins_anat.nii.gz")
        return outputs



from swane.utils.fsl_conflict_handler import fsl_conflict_check


def main():
    import sys
    import os
    from swane import strings
    import swane_supplement
    from PySide6.QtWidgets import QApplication, QMessageBox
    from PySide6.QtGui import QIcon, QPixmap
    from swane.ui.MainWindow import MainWindow
    from swane.config.ConfigManager import ConfigManager
    from swane import EXIT_CODE_REBOOT
    from swane.config.config_enums import GlobalPrefCategoryList
    from swane.utils.last_pid_is_running import last_pid_is_running

    # Exit Code definition for automatic reboot
    current_exit_code = EXIT_CODE_REBOOT

    while current_exit_code == EXIT_CODE_REBOOT:

        # Singleton for SWANe application
        if not QApplication.instance():
            app = QApplication(sys.argv)
        else:
            app = QApplication.instance()

        # SWANe Icon definition
        app.setWindowIcon(QIcon(QPixmap(swane_supplement.appIcon_file)))
        # SWANe App Name definition
        app.setApplicationDisplayName(strings.APPNAME)

        # SWANe Configuration loading
        global_config = ConfigManager()

        # Guard to prevent multiple SWANe instances launch
        last_pid = global_config.get_last_pid()

        if last_pid_is_running(last_pid):
            msg_box = QMessageBox()
            msg_box.setText(strings.main_multiple_instances_error)
            msg_box.exec()
            sys.exit(-1)
        else:
            global_config[GlobalPrefCategoryList.MAIN]["last_pid"] = str(os.getpid())
            global_config.save()

        # MainWindow in a varariable to prenvent garbage collector deletion (might cause crash)
        widget = MainWindow(global_config)
        widget.setWindowIcon(QIcon(QPixmap(swane_supplement.appIcon_file)))
        current_exit_code = app.exec()

    sys.exit(current_exit_code)


if __name__ == "__main__":

    # Before GUI execution check for fsl/python/freesurfer error
    if fsl_conflict_check():
        main()



# slicerpython script for module checking
# Warning: slicer library is not required beacuse the script is executed in Slicer environment

import sys

# Check FreeSurfer imported module into 3D Slicer
if hasattr(slicer.moduleNames, "FreeSurferImporter"):
    print("MODULE FOUND")
sys.exit(0)



import configparser
from swane import strings, __version__
from swane.config.preference_list import *
from swane.utils.CryptographyManager import CryptographyManager
from swane.utils.DataInputList import DataInputList
from enum import Enum
from swane.config.config_enums import WORKFLOW_TYPES, GlobalPrefCategoryList
from swane.utils.MailManager import MailManager


class ConfigManager(configparser.ConfigParser):

    # Overrides to accept non-str stringable object as section keys
    def __getitem__(self, key):
        return super().__getitem__(str(key))

    def __setitem__(self, key, value):
        super().__setitem__(str(key), value)

    def __init__(self, subject_folder: str = None, global_base_folder: str = None):
        """
        Parameters
        ----------
        subject_folder: path
            The subject folder path. None in global all configuration
        global_base_folder: path, optional
            An alternative folder for global configuration file. Default is None
        """
        super(ConfigManager, self).__init__()
        self._section_defaults = {}

        # First set some internal values differentiating global from subject pref objects
        if subject_folder is not None:
            self.global_config = False
            self.config_file = os.path.abspath(os.path.join(subject_folder, ".config"))
        else:
            if global_base_folder is None or not os.path.exists(global_base_folder):
                global_base_folder = os.path.expanduser("~")
            self.global_config = True
            self.config_file = os.path.abspath(
                os.path.join(global_base_folder, "." + strings.APPNAME)
            )

        # Load default pref from pref list
        self._load_defaults(save=False)

        # check if this version need pref reset
        force_pref_reset = self.getboolean_safe(
            GlobalPrefCategoryList.MAIN, "force_pref_reset"
        )

        reset_pref = False

        # if version need pref reset, load old config file in a temp variable to get just last_swane_version
        try:
            if force_pref_reset:
                if self.global_config:
                    last_swane_version = self[GlobalPrefCategoryList.MAIN][
                        "last_swane_version"
                    ]
                else:
                    temp_config = configparser.ConfigParser()
                    temp_config.read(self.config_file)
                    last_swane_version = temp_config[GlobalPrefCategoryList.MAIN][
                        "last_swane_version"
                    ]
                if __version__ != last_swane_version:
                    reset_pref = True
        except:
            pass

        if not reset_pref and os.path.exists(self.config_file):
            self.read(self.config_file)
            # Cycle all read values and reassign them to invoke validate_type without rewriting read method
            for section in self._section_defaults.keys():
                for option in self._section_defaults[section].keys():
                    if self._section_defaults[section][option].default_at_startup:
                        self.set(
                            section,
                            option,
                            self._section_defaults[section][option].default,
                        )
                    else:
                        self.set(section, option, self[section][option])

            # set last_swane_version in subject config to ensure force_pref_reset compatibility
            if str(GlobalPrefCategoryList.MAIN) not in self:
                self[GlobalPrefCategoryList.MAIN] = {}
            self[GlobalPrefCategoryList.MAIN]["last_swane_version"] = __version__

        self.save()

    def reload(self):
        """
        Reload the configuration file
        """
        self.read(self.config_file)

    def reset_to_defaults(self):
        """
        Discard current values and reload defaults
        """
        self._load_defaults(save=True)

    def _load_defaults(self, save: bool):
        """
        Loop the default dicts and insert default values in this object

        Parameters
        ----------
        save: bool
            If True, save the preferences
        """
        if self.global_config:
            for category in GlobalPrefCategoryList:
                if not save:
                    self[category] = {}
                    self._section_defaults[str(category)] = GLOBAL_PREFERENCES[category]
                    for pref in GLOBAL_PREFERENCES[category]:
                        if isinstance(GLOBAL_PREFERENCES[category][pref].default, Enum):
                            self[category][pref] = GLOBAL_PREFERENCES[category][
                                pref
                            ].default.name
                        else:
                            self[category][pref] = str(
                                GLOBAL_PREFERENCES[category][pref].default
                            )

            for data_input in DataInputList:
                if data_input in WF_PREFERENCES:
                    self[data_input] = {}
                    self._section_defaults[str(data_input)] = WF_PREFERENCES[data_input]
                    for pref in WF_PREFERENCES[data_input]:
                        if isinstance(WF_PREFERENCES[data_input][pref].default, Enum):
                            self[data_input][pref] = WF_PREFERENCES[data_input][
                                pref
                            ].default.name
                        else:
                            self[data_input][pref] = str(
                                WF_PREFERENCES[data_input][pref].default
                            )
        else:
            tmp_config = ConfigManager()
            for data_input in DataInputList:
                if data_input in WF_PREFERENCES:
                    self._section_defaults[str(data_input)] = WF_PREFERENCES[data_input]
                    self[data_input] = tmp_config[data_input]
            self[GlobalPrefCategoryList.MAIN] = {}
            self[GlobalPrefCategoryList.MAIN]["last_swane_version"] = tmp_config[
                GlobalPrefCategoryList.MAIN
            ]["last_swane_version"]
            self[GlobalPrefCategoryList.MAIN]["force_pref_reset"] = tmp_config[
                GlobalPrefCategoryList.MAIN
            ]["force_pref_reset"]

            self.set_workflow_option(
                tmp_config.getenum_safe(GlobalPrefCategoryList.MAIN, "default_wf_type")
            )
        if save:
            self.save()

    def set_workflow_option(self, workflow_type: WORKFLOW_TYPES):
        """
        Apply a workflow_type preset

        Parameters
        ----------
        workflow_type: WORKFLOW_TYPES
            The preset to apply
        """
        if self.global_config:
            return
        if type(workflow_type) is not WORKFLOW_TYPES:
            return
        self[DataInputList.T13D]["wf_type"] = workflow_type.name
        for category in DEFAULT_WF[workflow_type]:
            for key in DEFAULT_WF[workflow_type][category]:
                self[category][key] = DEFAULT_WF[workflow_type][category][key]

    def save(self):
        """
        Save the current preferences to the config file
        """
        with open(self.config_file, "w") as openedFile:
            self.write(openedFile)

    def get_main_working_directory(self) -> str:
        """
        Returns
        -------
        A string containing the main working directory
        """
        try:
            if self.global_config and os.path.exists(
                self[GlobalPrefCategoryList.MAIN]["main_working_directory"]
            ):
                return self[GlobalPrefCategoryList.MAIN]["main_working_directory"]
        except:
            pass
        return ""

    def set_main_working_directory(self, main_working_dir_path: str):
        """
        Set the main working directory

        Parameters
        ----------
        main_working_dir_path: path
            the new path

        """
        if self.global_config:
            self[GlobalPrefCategoryList.MAIN][
                "main_working_directory"
            ] = main_working_dir_path
            self.save()

    def get_max_subject_tabs(self) -> int:
        """
        Returns
        -------
        An int containing the maximum subject tab number
        """
        return self.getint_safe(GlobalPrefCategoryList.PERFORMANCE, "max_subj")

    def get_subjects_prefix(self) -> str:
        """
        Returns
        -------
        A string containing the subject folder prefix
        """
        if self.global_config:
            return self[GlobalPrefCategoryList.MAIN]["subjects_prefix"]
        return ""

    def get_default_dicom_folder(self) -> str:
        """
        Returns
        -------
        A string containing the default dicom folder name
        """
        if self.global_config:
            return self[GlobalPrefCategoryList.MAIN]["default_dicom_folder"]
        return ""

    def get_slicer_path(self) -> str:
        """
        Returns
        -------
        A string containing the slicer executable path
        """
        if self.global_config:
            return self[GlobalPrefCategoryList.MAIN]["slicer_path"]
        return ""

    def set_slicer_path(self, slicer_path: str):
        """
        Set the slicer executable path

        Parameters
        ----------
        slicer_path: path
            the new path

        """
        if self.global_config:
            self[GlobalPrefCategoryList.MAIN]["slicer_path"] = slicer_path

    def get_slicer_version(self) -> str:
        """
        Returns
        -------
        A string containing the current slicer version
        """
        if self.global_config:
            return self[GlobalPrefCategoryList.MAIN]["slicer_version"]

    def set_slicer_version(self, slicer_version: str):
        """
        Set the Slicer version

        Parameters
        ----------
        slicer_version: str
            the Slicer version

        """
        if self.global_config:
            self[GlobalPrefCategoryList.MAIN]["slicer_version"] = slicer_version

    def is_optional_series_enabled(self, series_name: DataInputList) -> bool:
        """
        Check if an optional series is enabled

        Parameters
        ----------
        series_name: DataInputList
            The series to check

        Returns
        -------
        True if the specified series is enabled
        """
        return self.getboolean_safe(
            GlobalPrefCategoryList.OPTIONAL_SERIES, str(series_name)
        )

    def get_slicer_scene_ext(self) -> str:
        """
        Returns
        -------
        A string containing the default slicer scene extension
        """
        if self.global_config:
            return self.getenum_safe(
                GlobalPrefCategoryList.MAIN, "slicer_scene_ext"
            ).value
        return None

    def get_subject_workflow_type(self) -> Enum:
        """
        Returns
        -------
        The Enum of default workflow type
        """
        return self.getenum_safe(DataInputList.T13D, "wf_type")

    def get_workflow_hippo_pref(self) -> bool:
        """
        Returns
        -------
        True if hippocampal segmentation is enabled
        """
        return self.getboolean_safe(DataInputList.T13D, "hippo_amyg_labels")

    def get_workflow_freesurfer_pref(self) -> bool:
        """
        Returns
        -------
        True if freesurfer analysis is enabled
        """
        return self.getboolean_safe(DataInputList.T13D, "freesurfer")

    def get_mail_manager(self) -> MailManager:
        """
        Returns
        -------
        An initialized MailManager
        """

        mail_manager_enabled = self.getboolean_safe(
            GlobalPrefCategoryList.MAIL_SETTINGS, "enabled"
        )
        if not mail_manager_enabled:
            return None

        server_address = self[GlobalPrefCategoryList.MAIL_SETTINGS]["address"]
        server_port = self.getint_safe(GlobalPrefCategoryList.MAIL_SETTINGS, "port")
        username = self[GlobalPrefCategoryList.MAIL_SETTINGS]["username"]
        password = CryptographyManager.decrypt(
            self[GlobalPrefCategoryList.MAIL_SETTINGS]["password"]
        )
        use_ssl = self.getboolean_safe(GlobalPrefCategoryList.MAIL_SETTINGS, "use_ssl")
        use_tls = self.getboolean_safe(GlobalPrefCategoryList.MAIL_SETTINGS, "use_tls")

        if (
            server_address == ""
            or server_port == ""
            or username == ""
            or password == ""
        ):
            return None

        mail_manager = MailManager(
            server_address=server_address,
            server_port=server_port,
            username=username,
            password=password,
            use_ssl=use_ssl,
            use_tls=use_tls,
        )

        return mail_manager

    def check_dependencies(self, dependency_manager):
        """
        Check if dependencies of each preference entry are met, otherwise set them to false

        Parameters
        ----------
        dependency_manager: DependencyManager
            The application dependencies

        """
        changed = False
        for category in WF_PREFERENCES:
            for key in WF_PREFERENCES[category]:
                if WF_PREFERENCES[category][key].dependency is not None:
                    dep_check = getattr(
                        dependency_manager,
                        WF_PREFERENCES[category][key].dependency,
                        None,
                    )
                    if dep_check is None or not callable(dep_check) or not dep_check():
                        self[category][key] = "false"
                        changed = True
        if changed:
            self.save()

    def get_last_pid(self) -> int:
        """
        Returns
        -------
        The last application process ID as an int
        """
        return self.getint_safe(GlobalPrefCategoryList.MAIN, "last_pid")

    def getboolean_safe(
        self, section: str, option: str, *, raw=False, vars=None, **kwargs
    ) -> bool:
        """
        Get an option value as bool or, if invalid, its default value

        Parameters
        ----------
        section: str
            The configuration section
        option: str
            The configuration option
        raw: bool, optional
            Unused, passed to super(). Default is False
        vars
            Unused, passed to Super(). Default is None

        Returns
        -------
        The option values bool
        """
        section = str(section)
        try:
            return self.getboolean(section, option, raw=raw, vars=vars)
        except:
            if (
                section in self._section_defaults
                and option in self._section_defaults[section]
            ):
                if type(self._section_defaults[section]) is list:
                    ret = self._section_defaults[section].default[0]
                else:
                    ret = self._section_defaults[section].default
                if ret.lower() in configparser.ConfigParser.BOOLEAN_STATES:
                    return configparser.ConfigParser.BOOLEAN_STATES[ret.lower()]
        raise Exception()

    def getint_safe(
        self, section: str, option: str, *, raw=False, vars=None, **kwargs
    ) -> int:
        """
        Get an option value as int or, if invalid, its default value

        Parameters
        ----------
        section: str
            The configuration section
        option: str
            The configuration option
        raw: bool, optional
            Unused, passed to super(). Default is False
        vars
            Unused, passed to Super(). Default is None

        Returns
        -------
        The option values int
        """
        section = str(section)
        try:
            return self.getint(section, option, raw=raw, vars=vars)
        except:
            if (
                section in self._section_defaults
                and option in self._section_defaults[section]
            ):
                if type(self._section_defaults[section][option].default) is list:
                    return 0
                else:
                    return int(self._section_defaults[section][option].default)
        raise Exception("Error for %s - %s" % (str(section), str(option)))

    def getfloat_safe(
        self, section: str, option: str, *, raw=False, vars=None, **kwargs
    ) -> float:
        """
        Get an option value as float or, if invalid, its default value

        Parameters
        ----------
        section: str
            The configuration section
        option: str
            The configuration option
        raw: bool, optional
            Unused, passed to super(). Default is False
        vars
            Unused, passed to Super(). Default is None

        Returns
        -------
        The option values as float
        """
        section = str(section)
        try:
            return self.getfloat(section, option, raw=raw, vars=vars)
        except:
            if (
                section in self._section_defaults
                and option in self._section_defaults[section]
            ):
                if type(self._section_defaults[section][option].default) is list:
                    return float(self._section_defaults[section][option].default[0])
                else:
                    return float(self._section_defaults[section][option].default)
        raise Exception("Error for %s - %s" % (str(section), str(option)))

    def getenum_safe(
        self, section: str, option: str, *, raw: bool = False, vars=None, **kwargs
    ) -> Enum:
        """
        Get an option value as Enum or, if invalid, its default value

        Parameters
        ----------
        section: str
            The configuration section
        option: str
            The configuration option
        raw: bool, optional
            Unused, passed to super(). Default is False
        vars
            Unused, passed to Super(). Default is None

        Returns
        -------
        The option values as Enum
        """
        section = str(section)
        if self._section_defaults[section][option].value_enum is None:
            raise Exception("No value_enum for %s - %s" % (str(section), str(option)))

        if (
            self[section][option]
            in self._section_defaults[section][option].value_enum.__members__
        ):
            return self._section_defaults[section][option].value_enum[
                self[section][option]
            ]
        else:
            return self._section_defaults[section][option].default

    def validate_type(self, section: str = "", option: str = "", value: str = ""):
        """
        Set an option after checking for type

        Parameters
        ----------
        section: str
            The configuration section
        option: str
            The configuration option
        value: str, optional
            The value to be checked. DEfault is None

        Returns
        -------
        The value cast to its type or the default option value
        """
        if section != "" and option != "" and value != "":
            if (
                section in self._section_defaults
                and option in self._section_defaults[section]
            ):
                if self._section_defaults[section][option].input_type == InputTypes.INT:
                    try:
                        return int(value)
                    except:
                        return int(self._section_defaults[section][option].default)
                elif (
                    self._section_defaults[section][option].input_type
                    == InputTypes.ENUM
                ):
                    if (
                        value
                        in self._section_defaults[section][
                            option
                        ].value_enum.__members__
                    ):
                        return value
                    else:
                        return self._section_defaults[section][option].default.name
                elif (
                    self._section_defaults[section][option].input_type
                    == InputTypes.FLOAT
                ):
                    try:
                        return float(value)
                    except:
                        return float(self._section_defaults[section][option].default)
                elif (
                    self._section_defaults[section][option].input_type
                    == InputTypes.BOOLEAN
                ):
                    if value.lower() in configparser.ConfigParser.BOOLEAN_STATES:
                        return value
                    elif (
                        self._section_defaults[section][option].default.lower()
                        in configparser.ConfigParser.BOOLEAN_STATES
                    ):
                        return self._section_defaults[section][option].default.lower()
        return value

    def set(self, section: str, option: str, value: str = None):
        """
        Set an option after checking for type

        Parameters
        ----------
        section: str
            The configuration section
        option: str
            The configuration option
        value: str, optional
            The value to be checked. DEfault is None

        Returns
        -------
        True if the application preference has a valide Slicer path
        """
        if value is not None:
            value = str(self.validate_type(section, option, value))
        super().set(section, option, value)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.base import (
    traits,
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
    Bunch,
    isdefined,
)
from swane.config.config_enums import BLOCK_DESIGN


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class FMRIGenSpecInputSpec(BaseInterfaceInputSpec):
    TR = traits.Float(mandatory=True, desc="Repetition Time")
    nvols = traits.Int(mandatory=True, desc="Number of EPI runs")
    task_duration = traits.Int(mandatory=True, desc="Task duration")
    rest_duration = traits.Int(mandatory=True, desc="Rest duration")
    block_design = traits.Enum(BLOCK_DESIGN, usedefault=True)
    task_a_name = traits.String(mandatory=False, desc="Task A name")
    task_b_name = traits.String(mandatory=False, desc="Task A name")
    hpcutoff = traits.Int(mandatory=False, desc="Cutoff for highpass filtering")
    tempMean = File(exists=True, mandatory=True, desc="mean functional image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class FMRIGenSpecOutputSpec(TraitedSpec):
    hpcutoff = traits.Int(desc="Cutoff for highpass filtering")
    hp_sigma_vol = traits.Float(desc="Sigma volume for highpass filtering")
    evs_run = traits.Any(desc="task events")
    hpstring = traits.String(desc="op_string for highpass filtering")
    task_a_name = traits.String(desc="Task name")
    task_b_name = traits.String(desc="Task name")
    contrasts = traits.List(desc="T contrast array")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class FMRIGenSpec(BaseInterface):
    """
    Formats the fMRI parameters for FSL feat nodes.

    """

    input_spec = FMRIGenSpecInputSpec
    output_spec = FMRIGenSpecOutputSpec

    def _run_interface(self, runtime):

        if isdefined(self.inputs.hpcutoff):
            self.hpcutoff = self.inputs.hpcutoff
        elif self.inputs.block_design == BLOCK_DESIGN.RARA:
            self.hpcutoff = self.inputs.task_duration + self.inputs.rest_duration
        else:
            self.hpcutoff = (self.inputs.task_duration + self.inputs.rest_duration) * 2

        self.hp_sigma_vol = self.hpcutoff / (2 * self.inputs.TR)

        self.hpstring = "-bptf %f -1 -add %s" % (
            self.hp_sigma_vol,
            self.inputs.tempMean,
        )

        if not isdefined(self.inputs.task_a_name):
            self.inputs.task_a_name = "TaskA"
        if not isdefined(self.inputs.task_b_name):
            self.inputs.task_b_name = "TaskB"

        if self.inputs.block_design == BLOCK_DESIGN.RARA:
            self.contrasts = [
                [
                    "%s_versus_Rest" % self.inputs.task_a_name,
                    "T",
                    [self.inputs.task_a_name],
                    [1],
                ]
            ]

            self.evs_run = Bunch(
                conditions=[self.inputs.task_a_name],
                onsets=[
                    list(
                        range(
                            self.inputs.rest_duration,
                            int(self.inputs.TR * self.inputs.nvols),
                            (self.inputs.task_duration + self.inputs.rest_duration),
                        )
                    )
                ],
                durations=[[self.inputs.task_duration]],
            )
        else:
            cont1 = [
                "%s_versus_%s" % (self.inputs.task_a_name, self.inputs.task_b_name),
                "T",
                [self.inputs.task_a_name, self.inputs.task_b_name],
                [1, -1],
            ]
            cont2 = [
                "%s_versus_%s" % (self.inputs.task_b_name, self.inputs.task_a_name),
                "T",
                [self.inputs.task_a_name, self.inputs.task_b_name],
                [-1, 1],
            ]
            self.contrasts = [cont1, cont2]

            self.evs_run = Bunch(
                conditions=[self.inputs.task_a_name, self.inputs.task_b_name],
                onsets=[
                    list(
                        range(
                            self.inputs.rest_duration,
                            int(self.inputs.TR * self.inputs.nvols),
                            (
                                self.inputs.task_duration * 2
                                + self.inputs.rest_duration * 2
                            ),
                        )
                    ),
                    list(
                        range(
                            (self.inputs.rest_duration * 2 + self.inputs.task_duration),
                            int(self.inputs.TR * self.inputs.nvols),
                            (
                                self.inputs.task_duration * 2
                                + self.inputs.rest_duration * 2
                            ),
                        )
                    ),
                ],
                durations=[[self.inputs.task_duration], [self.inputs.task_duration]],
            )

        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["hpcutoff"] = self.hpcutoff
        outputs["hp_sigma_vol"] = self.hp_sigma_vol
        outputs["hpstring"] = self.hpstring
        outputs["evs_run"] = self.evs_run
        outputs["contrasts"] = self.contrasts
        outputs["task_a_name"] = self.inputs.task_a_name
        outputs["task_b_name"] = self.inputs.task_b_name
        return outputs



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from os.path import abspath
import os

from nipype.interfaces.fsl import BinaryMaths
from nipype.interfaces.base import (
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    InputMultiPath,
    File,
    isdefined,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class SumMultiTracksInputSpec(BaseInterfaceInputSpec):
    path_files = InputMultiPath(
        File(exists=True), mandatory=True, desc="list of path file to sum togheter"
    )
    waytotal_files = InputMultiPath(
        File(exists=True), mandatory=True, desc="list of waytotal files to sum togheter"
    )
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class SumMultiTracksOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output image")
    waytotal_sum = File(exists=True, desc="the output waytotal file")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class SumMultiTracks(BaseInterface):
    """
    Merges results from multiple tractography runs.

    """

    input_spec = SumMultiTracksInputSpec
    output_spec = SumMultiTracksOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file = self._gen_outfilename()
        waytotal_sum_file = self._gen_waytotal_outfilename()

        steps = len(self.inputs.path_files)
        sum_loop = [None] * steps
        sum_res = [None] * steps
        waytotal_sum = 0

        for x in range(steps):

            # SUM FTP_PATHS
            sum_loop[x] = BinaryMaths()
            sum_loop[x].inputs.operation = "add"

            if x == 0:
                sum_loop[x].inputs.in_file = self.inputs.path_files[x]
            else:
                sum_loop[x].inputs.in_file = sum_res[(x - 1)].outputs.out_file

            sum_loop[x].inputs.operand_file = self.inputs.path_files[x]

            if x == (steps - 1):
                sum_loop[x].inputs.out_file = self.inputs.out_file

            sum_res[x] = sum_loop[x].run()

            # SUM WAYTOTAL
            if os.path.exists(self.inputs.waytotal_files[x]):
                with open(self.inputs.waytotal_files[x], "r") as file:
                    for line in file.readlines():
                        waytotal_sum += int(line)

        with open(waytotal_sum_file, "w") as file:
            file.write(str(waytotal_sum))

        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = "sum.nii.gz"
        return abspath(out_file)

    def _gen_waytotal_outfilename(self):
        out_file = os.path.basename(self.inputs.out_file)
        if not isdefined(out_file):
            out_file = "waytotal"
        else:
            out_file = out_file.replace(".nii.gz", "") + "_waytotal"
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        outputs["waytotal_sum"] = self._gen_waytotal_outfilename()
        return outputs



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl.base import FSLCommand, FSLCommandInputSpec
from nipype.interfaces.base import traits, TraitedSpec, File, isdefined


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.base.FSLCommandInputSpec)  -*-
class GetNiftiTRInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s pixdim4",
        position="1",
        desc="the input image",
    )
    force_value = traits.Float(mandatory=False, desc="value forced by user")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class GetNiftiTROutputSpec(TraitedSpec):
    TR = traits.Float(desc="Repetition Time")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.base.FSLCommand)  -*-
class GetNiftiTR(FSLCommand):
    """
    Reads the time of repetition from a NIFTI file.

    """

    _cmd = "fslval"
    input_spec = GetNiftiTRInputSpec
    output_spec = GetNiftiTROutputSpec

    def aggregate_outputs(self, runtime=None, needed_outputs=None):
        outputs = self._outputs()

        # se l'utente ha inserito un valore forzo quello invece del risultato della lettura automatica
        if isdefined(self.inputs.force_value) and self.inputs.force_value != -1:
            outputs.TR = self.inputs.force_value
            return outputs

        info = runtime.stdout
        try:
            outputs.TR = float(info)
        except:
            outputs.TR = 0.0

        return outputs
